@@ -5,7 +5,7 @@
  * Copyright (c) 2000-2009 Ethan Galstad (egalstad@nagios.org)
  * License: GPL v2
  *
- * Last Modified: 2020-04-15
+ * Last Modified: 2021-10-27
  *
  * Command line: NSCA -c <config_file> [mode]
  *
@@ -37,6 +37,7 @@ static char password[MAX_INPUT_BUFFER]="";
 static enum { OPTIONS_ERROR, SINGLE_PROCESS_DAEMON, MULTI_PROCESS_DAEMON, INETD } mode=SINGLE_PROCESS_DAEMON;
 static int foreground=FALSE;
 static int debug=FALSE;
+static int strict_mode_spoofing=FALSE;
 static int aggregate_writes=FALSE;
 static int decryption_method=ENCRYPT_XOR;
 static int append_to_file=FALSE;
@@ -477,6 +478,14 @@ static int read_config_file(char *filename){
                         else
                                 debug=FALSE;
                         }
+        else if (strstr(input_buffer, "strict_mode_spoofing")) {
+                        if (atoi(varvalue) > 0) {
+                            strict_mode_spoofing = TRUE;
+                        }
+                        else {
+                            strict_mode_spoofing = FALSE;
+                        }
+        }
 		else if(strstr(input_buffer,"aggregate_writes")){
                         if(atoi(varvalue)>0)
                                 aggregate_writes=TRUE;
@@ -926,35 +935,38 @@ static void wait_for_connections(void) {
 
 
 
-static void accept_connection(int sock, void *unused){
-        int new_sd;
-        pid_t pid;
-        struct sockaddr_storage addr;
-        socklen_t addrlen;
-        char hostbuf[64], portbuf[16];
-        char *h;
-        int rc;
+static void accept_connection(int sock, void *unused) {
+    int new_sd;
+    pid_t pid;
+    struct sockaddr_storage addr;
+    socklen_t addrlen;
+    char hostbuf[64], portbuf[16];
+    char *h;
+    int rc;
 #ifdef HAVE_LIBWRAP
 	struct request_info req;
 #endif
 
 	/* DO NOT REMOVE! 01/29/2007 single process daemon will fail if this is removed */
-        if(mode==SINGLE_PROCESS_DAEMON)
-                register_read_handler(sock,accept_connection,NULL);
+    if(mode==SINGLE_PROCESS_DAEMON) {
+        register_read_handler(sock,accept_connection,NULL);
+    }
 
-        /* wait for a connection request */
-        while(1){
+    /* wait for a connection request */
+    while(1) {
 
-		/* we got a live one... */
-                if((new_sd=accept(sock,0,0))>=0)
-                        break;
+	/* we got a live one... */
+        if((new_sd=accept(sock,0,0))>=0) {
+            break;
+        }
 
-		/* handle the error */
-		else{
+	/* handle the error */
+	    else {
 
 			/* bail out if necessary */
-			if(sigrestart==TRUE || sigshutdown==TRUE)
+			if(sigrestart==TRUE || sigshutdown==TRUE) {
 				return;
+            }
 
 			/* try and handle temporary errors */
 			if(errno==EWOULDBLOCK || errno==EINTR || errno==ECHILD || errno==ECONNABORTED){
@@ -962,24 +974,26 @@ static void accept_connection(int sock, void *unused){
 					sleep(1);
 				else
 					return;
-				}
-			else
+		    }
+			else {
 				break;
-			}
-                }
+            }
+		}
+    }
 
-        /* hey, there was an error... */
-        if(new_sd<0){
+    /* hey, there was an error... */
+    if(new_sd<0){
 
-                /* log error to syslog facility */
-                syslog(LOG_ERR,"Network server accept failure (%d: %s)",errno,strerror(errno));
+            /* log error to syslog facility */
+            syslog(LOG_ERR,"Network server accept failure (%d: %s)",errno,strerror(errno));
 
-                /* close socket prior to exiting */
-                close(sock);
-		if(mode==MULTI_PROCESS_DAEMON)
+            /* close socket prior to exiting */
+            close(sock);
+		if(mode==MULTI_PROCESS_DAEMON) {
 			do_exit(STATE_CRITICAL);
+        }
 		return;
-                }
+    }
 
 #ifdef HAVE_LIBWRAP
 
@@ -992,60 +1006,63 @@ static void accept_connection(int sock, void *unused){
 		syslog(LOG_ERR, "refused connect from %s", eval_client(&req));
 		close(new_sd);
 		return;
-		}
+	}
 #endif
 
 
-        /* fork() if we have to... */
-        if(mode==MULTI_PROCESS_DAEMON){
+    /* fork() if we have to... */
+    if(mode==MULTI_PROCESS_DAEMON){
 
-                pid=fork();
-                if(pid){
-                        /* parent doesn't need the new connection */
-                        close(new_sd);
-                        return;
-                        }
-		else{
-                        /* child does not need to listen for connections */
-                        close(sock);
-                        }
-                }
+        pid=fork();
+        if(pid) {
+            /* parent doesn't need the new connection */
+            close(new_sd);
+            return;
+        }
+        else{
+            /* child does not need to listen for connections */
+            close(sock);
+        }
+    }
 
-        /* find out who just connected... */
-        addrlen=sizeof(addr);
-        rc=getpeername(new_sd,(struct sockaddr *)&addr,&addrlen);
+    /* find out who just connected... */
+    addrlen=sizeof(addr);
+    rc=getpeername(new_sd,(struct sockaddr *)&addr,&addrlen);
 
-        if(rc<0){
-                /* log error to syslog facility */
-                syslog(LOG_ERR,"Error: Network server getpeername() failure (%d: %s)",errno,strerror(errno));
+    if(rc<0){
+        /* log error to syslog facility */
+        syslog(LOG_ERR,"Error: Network server getpeername() failure (%d: %s)",errno,strerror(errno));
 
-                /* close socket prior to exiting */
-                close(new_sd);
-		if(mode==MULTI_PROCESS_DAEMON)
+        /* close socket prior to exiting */
+        close(new_sd);
+		if(mode==MULTI_PROCESS_DAEMON) {
 			do_exit(STATE_CRITICAL);
+        }
 		return;
-                }
+    }
 
-        /* log info to syslog facility */
-        if(debug==TRUE) {
-                getnameinfo((struct sockaddr *)&addr, addrlen,
-                        hostbuf, sizeof(hostbuf),
-                        portbuf, sizeof(portbuf),
-                        NI_NUMERICHOST|NI_NUMERICSERV);
-		h = strncmp(hostbuf, "::ffff:", 7) == 0 ? hostbuf + 7 : hostbuf;
-                syslog(LOG_DEBUG,"Connection from %s port %s",h,portbuf);
-                }
+    /* log info to syslog facility */
+    if(debug==TRUE) {
+        getnameinfo((struct sockaddr *)&addr, addrlen,
+                    hostbuf, sizeof(hostbuf),
+                    portbuf, sizeof(portbuf),
+                    NI_NUMERICHOST|NI_NUMERICSERV);
+	    h = strncmp(hostbuf, "::ffff:", 7) == 0 ? hostbuf + 7 : hostbuf;
+        syslog(LOG_DEBUG,"Connection from %s port %s",h,portbuf);
+    }
 
 	/* handle the connection */
-	if(mode==SINGLE_PROCESS_DAEMON)
+	if(mode==SINGLE_PROCESS_DAEMON) {
 		/* mark the connection as ready to be handled */
 		register_write_handler(new_sd, handle_connection, NULL);
-	else
+    }
+	else {
 		/* handle the client connection */
 		handle_connection(new_sd, NULL);
+    }
 
 	return;
-        }
+}
 
 
 
@@ -1124,7 +1141,146 @@ static void handle_connection(int sock, void *data){
 	return;
         }
 
+/* Takes the peer socket and the host_name that the peer is claiming for a check result.
+ * Returns TRUE iff the peer socket's address matches one of the host_name's addresses
+ * according to getaddrinfo().
+ */
+static int strict_mode_verify_spoofing(int sock, char *host_name)
+{
 
+    // Retrieve the address associated with the socket fd
+
+    /* Note: we did run getpeername() earlier in the program, but adding
+     * parameters and passing data around is difficult due to the event 
+     * processing code. (Search for 'rhand' to see the relevant code.)
+     */ 
+
+    struct sockaddr_storage peer_addr;
+    int peer_addr_len;
+    int status;
+    peer_addr_len = sizeof(peer_addr);
+    status = getpeername(sock, (struct sockaddr *)&peer_addr, &peer_addr_len);
+    if (status == -1) {
+        char *errmsg = strerror(errno);
+        syslog(LOG_ERR, "Strict mode returning early - getpeername() failed: %s", errmsg);
+        return FALSE;
+    }
+    // Network-order bytes are in addr.sin_addr
+
+    // Retrieve the address associated withe the host name we just read
+    struct addrinfo hints, *ai;
+
+    memset(&hints, 0, sizeof(hints));
+    hints.ai_family = peer_addr.ss_family;
+    hints.ai_flags = AI_ADDRCONFIG | AI_PASSIVE;
+    hints.ai_socktype = SOCK_STREAM;
+    hints.ai_protocol = IPPROTO_TCP;
+
+    status = getaddrinfo(host_name, NULL, &hints, &ai);
+
+    // A hostname can have multiple addresses.
+    // We don't have port information, so we'll check all of them
+    for (; ai != NULL; ai = ai->ai_next) {
+        if (ai->ai_addr->sa_family != peer_addr.ss_family) {
+            // Should already be filtered, but we'll check for it anyways
+            continue;
+        }
+
+        if (ai->ai_addr->sa_family == AF_INET) {
+            struct sockaddr_in *peer_as_ipv4 = ((struct sockaddr_in *) &peer_addr);
+            struct sockaddr_in *claimed_as_ipv4 = ((struct sockaddr_in *) ai->ai_addr);
+            unsigned long peer_network_order = peer_as_ipv4->sin_addr.s_addr;
+            unsigned long claimed_network_order = claimed_as_ipv4->sin_addr.s_addr;
+
+            // Both addresses should be in network order, so just compare longs
+            if (peer_network_order == claimed_network_order) {
+                return TRUE;
+            }
+        }
+        else if (ai->ai_addr->sa_family == AF_INET6) {
+            struct sockaddr_in6 *peer_as_ipv6 = ((struct sockaddr_in6 *) &peer_addr);
+            struct sockaddr_in6 *claimed_as_ipv6 = ((struct sockaddr_in6 *) ai->ai_addr);
+            unsigned char *peer_network_order = peer_as_ipv6->sin6_addr.s6_addr;
+            unsigned char *claimed_network_order = claimed_as_ipv6->sin6_addr.s6_addr;
+
+            int ipv6_no_differences = TRUE;
+            int i;
+            for (i = 0; i < 16; ++i)
+            {
+                ipv6_no_differences &= peer_network_order[i] == claimed_network_order[i];
+            }
+
+            if (ipv6_no_differences) {
+                return TRUE;
+            }
+        }
+    }
+
+    return FALSE;
+} 
+
+/* Convert newlines into the literals '\' and 'n'. Nagios Core runs the inverse
+ * of this function in both check result files and in the external commands file.
+ */
+char *escape_newlines(const char *rawbuf) {
+    char *newbuf = NULL;
+    int x;
+    int y;
+
+    if (rawbuf == NULL)
+        return NULL;
+
+    /* Count the escapes we need to make. */
+    for (x = 0, y = 0; rawbuf[x]; x++) {
+        if (rawbuf[x] == '\\' || rawbuf[x] == '\n')
+            y++;
+        }
+
+    /* Just duplicate the string if we have nothing to escape. */
+    if (y == 0)
+        return strdup(rawbuf);
+
+    /* Allocate memory for the new string with escapes. */
+    if ((newbuf = malloc(x + y + 1)) == NULL)
+        return NULL;
+
+    for (x = 0, y = 0; rawbuf[x]; x++) {
+
+        /* Escape backslashes. */
+        if (rawbuf[x] == '\\') {
+            newbuf[y++] = '\\';
+            newbuf[y++] = '\\';
+            }
+
+        /* Escape newlines. */
+        else if (rawbuf[x] == '\n') {
+            newbuf[y++] = '\\';
+            newbuf[y++] = 'n';
+            }
+
+        else
+            newbuf[y++] = rawbuf[x];
+        }
+    newbuf[y] = '\0';
+
+    return newbuf;
+}
+
+/* If the condition in this loop is triggered, the input is already malformed 
+ * for the purposes of writing to the external commands file. Let's just get
+ * rid of whatever weird thing was happening there.
+ */
+static inline void truncate_newlines(char *to_strip, size_t maxlen)
+{
+    size_t i;
+    for (i = 0; i < maxlen && to_strip[i] != 0; ++i)
+    {
+        if (to_strip[i] == '\n')
+        {
+            to_strip[i] = 0;
+        }
+    }
+}
 
 /* handle reading from a client connection */
 static void handle_connection_read(int sock, void *data){
@@ -1250,6 +1406,18 @@ static void handle_connection_read(int sock, void *data){
 			syslog(LOG_NOTICE,"SERVICE CHECK -> Host Name: '%s', Service Description: '%s', Return Code: '%d', Output: '%s'",host_name,svc_description,return_code,plugin_output);
 	        }
 
+        if (strict_mode_spoofing) {
+
+            int found_match = strict_mode_verify_spoofing(sock, host_name);
+
+            // If they don't match, reject the message and log the interaction
+            if (found_match == FALSE) {
+                syslog(LOG_WARNING, "Strict mode - dropped check for %s due to non-matching host name.", host_name);
+                return;
+            }
+
+        }
+
         /* write the check result to the external command file.
          * Note: it's OK to hang at this point if the write doesn't succeed, as there's
          * no way we could handle any other connection properly anyway.  so we don't
@@ -1257,102 +1425,121 @@ static void handle_connection_read(int sock, void *data){
          * only ever write one command at a time into the pipe.
          */
         //syslog(LOG_ERR,"'%s' (%s) []",check_result_path, strlen(check_result_path));
+        truncate_newlines(host_name, MAX_HOSTNAME_LENGTH);
+        truncate_newlines(svc_description, MAX_DESCRIPTION_LENGTH);
+        char *plugin_output_escaped = escape_newlines(plugin_output);
+
+
         if (check_result_path==NULL){
-        write_check_result(host_name,svc_description,return_code,plugin_output,time(NULL));
+            write_check_result(host_name,svc_description,return_code,plugin_output_escaped,time(NULL));
         }else{
-                write_checkresult_file(host_name,svc_description,return_code,plugin_output,time(NULL));
+            write_checkresult_file(host_name,svc_description,return_code,plugin_output_escaped,time(NULL));
         }
 
+        free(plugin_output_escaped);
+
 	return;
-        }
+}
 
 
 
 /* writes service/host check results to the Nagios checkresult directory */
-static int write_checkresult_file(char *host_name, char *svc_description, int return_code, char *plugin_output, time_t check_time){
-	if(debug==TRUE)
-		syslog(LOG_ERR,"Attempting to write checkresult file");
-        mode_t new_umask=077;
-        mode_t old_umask;
-        time_t current_time;
-        int checkresult_file_fd=-1;
-        char *checkresult_file=NULL;
-        char *checkresult_ok_file=NULL;
-        FILE *checkresult_file_fp=NULL;
-        FILE *checkresult_ok_file_fp=NULL;
-        /* change and store umask */
-        old_umask=umask(new_umask);
+static int write_checkresult_file(char *host_name, char *svc_description, int return_code, char *plugin_output, time_t check_time)
+{
+    if(debug==TRUE) {
+        syslog(LOG_ERR,"Attempting to write checkresult file");
+    }
+    mode_t new_umask=077;
+    mode_t old_umask;
+    time_t current_time;
+    int checkresult_file_fd=-1;
+    char *checkresult_file=NULL;
+    char *checkresult_ok_file=NULL;
+    FILE *checkresult_file_fp=NULL;
+    FILE *checkresult_ok_file_fp=NULL;
+    /* change and store umask */
+    old_umask=umask(new_umask);
 
-        /* create safe checkresult file */
-        asprintf(&checkresult_file,"%s/cXXXXXX",check_result_path);
-        checkresult_file_fd=mkstemp(checkresult_file);
-        if(checkresult_file_fd>0){
-                checkresult_file_fp=fdopen(checkresult_file_fd,"w");
-        } else {
-                syslog(LOG_ERR,"Unable to open and write checkresult file '%s', failing back to PIPE",checkresult_file);
-                return write_check_result(host_name,svc_description,return_code,plugin_output,check_time);
-                }
+    /* create safe checkresult file */
+    asprintf(&checkresult_file,"%s/cXXXXXX",check_result_path);
+    checkresult_file_fd=mkstemp(checkresult_file);
+    if(checkresult_file_fd>0) {
+        checkresult_file_fp=fdopen(checkresult_file_fd,"w");
+    }
+    else {
+        syslog(LOG_ERR,"Unable to open and write checkresult file '%s', failing back to PIPE",checkresult_file);
+        return write_check_result(host_name,svc_description,return_code,plugin_output,check_time);
+    }
 
-	if(debug==TRUE)
-		syslog(LOG_ERR,"checkresult file '%s' open for write.",checkresult_file);
+    if(debug==TRUE) {
+        syslog(LOG_ERR,"checkresult file '%s' open for write.",checkresult_file);
+    }
 
-        time(&current_time);
-        fprintf(checkresult_file_fp,"### NSCA Passive Check Result ###\n");
-        fprintf(checkresult_file_fp,"# Time: %s",ctime(&current_time));
-        fprintf(checkresult_file_fp,"file_time=%ld\n\n",current_time);
-        fprintf(checkresult_file_fp,"### %s Check Result ###\n",(!*svc_description)?"Host":"Service");
-        fprintf(checkresult_file_fp,"host_name=%s\n",host_name);
-        if(strcmp(svc_description,""))
-                fprintf(checkresult_file_fp,"service_description=%s\n",svc_description);
-        fprintf(checkresult_file_fp,"check_type=1\n");
-        fprintf(checkresult_file_fp,"scheduled_check=0\n");
-        fprintf(checkresult_file_fp,"reschedule_check=0\n");
-        /* We have no latency data at this point. */
-        fprintf(checkresult_file_fp,"latency=0\n");
-        fprintf(checkresult_file_fp,"start_time=%lu.%lu\n",check_time,0L);
-        fprintf(checkresult_file_fp,"finish_time=%lu.%lu\n",check_time,0L);
-        fprintf(checkresult_file_fp,"return_code=%d\n",return_code);
-        /* newlines in output are already escaped */
-        fprintf(checkresult_file_fp,"output=%s\n",(plugin_output==NULL)?"":plugin_output);
-        fprintf(checkresult_file_fp,"\n");
+    time(&current_time);
+    fprintf(checkresult_file_fp,"### NSCA Passive Check Result ###\n");
+    fprintf(checkresult_file_fp,"# Time: %s",ctime(&current_time));
+    fprintf(checkresult_file_fp,"file_time=%ld\n\n",current_time);
+    fprintf(checkresult_file_fp,"### %s Check Result ###\n",(!*svc_description)?"Host":"Service");
+    fprintf(checkresult_file_fp,"host_name=%s\n",host_name);
+    if(strcmp(svc_description,"")) {
+        fprintf(checkresult_file_fp,"service_description=%s\n",svc_description);
+    }
+    fprintf(checkresult_file_fp,"check_type=1\n");
+    fprintf(checkresult_file_fp,"scheduled_check=0\n");
+    fprintf(checkresult_file_fp,"reschedule_check=0\n");
+    /* We have no latency data at this point. */
+    fprintf(checkresult_file_fp,"latency=0\n");
+    fprintf(checkresult_file_fp,"start_time=%lu.%lu\n",check_time,0L);
+    fprintf(checkresult_file_fp,"finish_time=%lu.%lu\n",check_time,0L);
+    fprintf(checkresult_file_fp,"return_code=%d\n",return_code);
+    /* newlines in output are already escaped */
+    fprintf(checkresult_file_fp,"output=%s\n",(plugin_output==NULL)?"":plugin_output);
+    fprintf(checkresult_file_fp,"\n");
 
-        fclose(checkresult_file_fp);
-        /* create and close ok file */
-        asprintf(&checkresult_ok_file,"%s.ok",checkresult_file);
+    fclose(checkresult_file_fp);
+    /* create and close ok file */
+    asprintf(&checkresult_ok_file,"%s.ok",checkresult_file);
+    if(debug==TRUE) {
         syslog(LOG_DEBUG,"checkresult completion file '%s' open.",checkresult_ok_file);
-        checkresult_ok_file_fp = fopen(checkresult_ok_file,"w");
-        fclose(checkresult_ok_file_fp);
-        /* reset umask */
-        umask(old_umask);
+    }
+    checkresult_ok_file_fp = fopen(checkresult_ok_file,"w");
+    fclose(checkresult_ok_file_fp);
+    /* reset umask */
+    umask(old_umask);
 
-        return OK;
-        }
+    return OK;
+}
 
 /* writes service/host check results to the Nagios command file */
-static int write_check_result(char *host_name, char *svc_description, int return_code, char *plugin_output, time_t check_time){
-	if(debug==TRUE)
+static int write_check_result(char *host_name, char *svc_description, int return_code, char *plugin_output, time_t check_time) {
+	if(debug==TRUE) {
 		syslog(LOG_ERR,"Attempting to write to nagios command pipe");
-        if(aggregate_writes==FALSE){
-                if(open_command_file()==ERROR)
-                        return ERROR;
-                }
+    }
+    if(aggregate_writes==FALSE){
+        if(open_command_file()==ERROR) {
+            return ERROR;
+        }
+    }
 
-	if(!strcmp(svc_description,""))
+	if(!strcmp(svc_description,"")) {
 		fprintf(command_file_fp,"[%lu] PROCESS_HOST_CHECK_RESULT;%s;%d;%s\n",(unsigned long)check_time,host_name,return_code,plugin_output);
+    }
 	else{
 		fprintf(command_file_fp,"[%lu] PROCESS_SERVICE_CHECK_RESULT;%s;%s;%d;%s\n",(unsigned long)check_time,host_name,svc_description,return_code,plugin_output);
-                }
-        if(aggregate_writes==FALSE)
-                close_command_file();
-        else
-                /* if we don't fflush() then we're writing in 4k non-CR-terminated blocks, and
-                 * anything else (eg. pscwatch) which writes to the file will be writing into
-                 * the middle of our commands.
-                 */
-                fflush(command_file_fp);
+    }
+    if(aggregate_writes==FALSE) {
+        close_command_file();
+    }
+    else {
+        /* if we don't fflush() then we're writing in 4k non-CR-terminated blocks, and
+         * anything else (eg. pscwatch) which writes to the file will be writing into
+         * the middle of our commands.
+         */
+        fflush(command_file_fp);
+    }
 
-        return OK;
-        }
+    return OK;
+}
 
 
 
