@@ -4,7 +4,7 @@
  * License: GPL v2
  * Copyright (c) 2000-2007 Ethan Galstad (nagios@nagios.org)
  *
- * Last Modified: 2020-04-15
+ * Last Modified: 2021-10-27
  *
  * Command line: SEND_NSCA <host_address> [-p port] [-to to_sec] [-c config_file]
  *
@@ -411,7 +411,29 @@ int read_init_packet(int sock){
 	return OK;
 }
 
+/* 
+ * Reads command-line argument arg and converts into a delimiter string, stored in result.
+ * For any single-character argument, the literal character is used as the argument.
+ * Otherwise, if a number is given, the argument will be converted to a number, and the 
+ * corresponding ASCII code will be used.
+ * e.g. if "9" is given, "9" will be used as a separator, but if "0x9" is given, the tab
+ * character ("\t") will be used instead.
+ */
+int parse_delimiter(char *result, size_t result_size, const char *arg) {
+	if (strlen(arg) > 1 && ((arg[0] > 47 && arg[0] < 58) || arg[0] == 43 || arg[0] == 45)) {
+		/* arg starts with 0-9, +, or -, but isn't a single character */
+		result[0] = (char) strtol(arg, NULL, 0);
+		if (errno) {
+			return ERROR;
+		}
+	}
+	else {
+	    snprintf(result,result_size,"%s",arg);
+	    delimiter[result_size-1]='\x0';
+	}
 
+	return OK;
+}
 
 /* process command line arguments */
 int process_arguments(int argc, char **argv){
@@ -496,34 +518,18 @@ int process_arguments(int argc, char **argv){
 
 		/* delimiter to use when parsing input */
 		else if(!strcmp(argv[x-1],"-d")){
-			if(x<argc){
-				errno=0;
-				long int d = strtol(argv[x], NULL, 16);
-				if(errno){
-				    snprintf(delimiter,sizeof(delimiter),"%s",argv[x]);
-				    delimiter[sizeof(delimiter)-1]='\x0';
-			        }else delimiter[0]= (char) d;
-				x++;
-				}
-			else {
+			if (parse_delimiter(delimiter, sizeof(delimiter), argv[x])) {
 				return ERROR;
 			}
+			x++;
 		}
 
 		/* delimiter to use when parsing input set */
 		else if(!strcmp(argv[x-1],"-ds")){
-			if(x<argc){
-				errno=0;
-				long int d = strtol(argv[x], NULL, 16);
-				if(errno){
-				    snprintf(block_delimiter,sizeof(block_delimiter),"%s",argv[x]);
-				    block_delimiter[sizeof(block_delimiter)-1]='\x0';
-			        }else block_delimiter[0]= (char) d;
-				x++;
-				}
-			else {
+			if (parse_delimiter(block_delimiter, sizeof(block_delimiter), argv[x])) {
 				return ERROR;
 			}
+			x++;
 		}
 
 		else if(x>2)
@@ -540,7 +546,7 @@ void alarm_handler(int sig){
 	char *msg = NULL;
 	asprintf(&msg, "Error: Timeout after %d seconds\n",socket_timeout);
 	/* fprintf(stderr, "Error: Timeout after %d seconds\n",socket_timeout); */
-	write(STDERR_FILENO, msg, sizeof(msg) - 1);
+	write(STDERR_FILENO, msg, strlen(msg));
 
 	do_exit(STATE_CRITICAL);
 }
